import json
import requests

from .ingest_auth_agent import IngestAuthAgent


class IngestUIAgent:

    INGEST_UI_URL_TEMPLATE = "http://ingest.{}.data.humancellatlas.org"

    def __init__(self, deployment):
        self.deployment = deployment
        self.ingest_broker_url = self.INGEST_UI_URL_TEMPLATE.format(self.deployment)
        self.ingest_auth_agent = IngestAuthAgent()
        self.auth_headers = self.ingest_auth_agent.make_auth_header()

    def upload(self, metadata_spreadsheet_path):
        url = self.ingest_broker_url + '/api_upload'
        files = {'file': open(metadata_spreadsheet_path, 'rb')}
        response = requests.post(url, files=files, allow_redirects=False, headers=self.auth_headers)
        if response.status_code != requests.codes.found and response.status_code != requests.codes.created:
            raise RuntimeError(f"POST {url} response was {response.status_code}: {response.content}")
        return json.loads(response.content)['details']['submission_id']



import json

from termcolor import colored


class Project:
    """
    Model an Ingest Project entity
    """

    @classmethod
    def load_by_id(cls, project_id, ingest_api_agent):
        data = ingest_api_agent.get(f"/projects/{project_id}")
        return Project(project_data=data, ingest_api_agent=ingest_api_agent)

    @classmethod
    def load_by_uuid(cls, project_uuid, ingest_api_agent):
        data = ingest_api_agent.get(f"/projects/search/findByUuid?uuid={project_uuid}")
        return Project(project_data=data, ingest_api_agent=ingest_api_agent)

    def __init__(self, project_data=None, ingest_api_agent=None):
        self.api = ingest_api_agent
        self.data = project_data

    def __str__(self, prefix=""):
        return colored(f"{prefix}Project {self.id}\n", 'green') + \
               f"{prefix}    uuid={self.uuid}\n" \
               f"{prefix}    short_name={self.short_name}"

    @property
    def id(self):
        return self.data['_links']['self']['href'].split('/')[-1]

    @property
    def uuid(self):
        return self.data['uuid']['uuid']

    @property
    def short_name(self):
        return self.data['content']['project_core']['project_short_name']

    def show_associated(self, entities_to_show, prefix="", verbose=False):
        if 'submissions' in entities_to_show:
            for subm in self.submission_envelopes():
                print(subm.__str__(prefix=prefix))

    def submission_envelopes(self):
        data = self.api.get(self.data['_links']['submissionEnvelopes']['href'])
        return [
            SubmissionEnvelope(submission_data=subm_data, ingest_api_agent=self.api)
            for subm_data in data['_embedded']['submissionEnvelopes']
        ]


class SubmissionEnvelope:
    """
    Model an Ingest Submission Envelope entity
    """

    @classmethod
    def load_by_id(cls, submission_id, ingest_api_agent):
        data = ingest_api_agent.get(f"/submissionEnvelopes/{submission_id}")
        return SubmissionEnvelope(submission_data=data, ingest_api_agent=ingest_api_agent)

    @classmethod
    def iter_submissions(cls, ingest_api_agent):
        for page in ingest_api_agent.iter_pages('/submissionEnvelopes', page_size=500, sort='submissionDate,desc'):
            for submission_data in page['submissionEnvelopes']:
                yield SubmissionEnvelope(submission_data=submission_data, ingest_api_agent=ingest_api_agent)

    def __init__(self, submission_data, ingest_api_agent):
        self.data = submission_data
        self.api = ingest_api_agent
        self.envelope_id = self.data['_links']['self']['href'].split('/')[-1]

    def __str__(self, prefix=""):
        return colored(f"{prefix}SubmissionEnvelope {self.envelope_id}\n", 'green') + \
               f"{prefix}    uuid={self.uuid}\n" \
               f"{prefix}    status={self.status}"

    def show_associated(self, entities_to_show, prefix="", verbose=False):
        self.verbose = verbose
        if 'bundles' in entities_to_show:
            if self.status == 'Complete':
                print(f"{prefix}Bundles:")
                for bundle in self.bundles():
                    print(prefix + "    " + bundle)

        if 'files' in entities_to_show:
                print(f"{prefix}Files:")
                for file in self.files():
                    self._output(file.__str__(prefix=prefix))

    def _output(self, thing):
        if self.verbose:
            print("%r" % thing)
        else:
            print(thing)

    def files(self):
        return [File(file_data) for file_data in
                self.api.get_all(self.data['_links']['files']['href'], 'files')]

    def iter_files(self):
        url = self.data['_links']['files']['href']
        for page in self.api.iter_pages(url):
            for file in page['files']:
                yield file

    def projects(self):
        return [Project(project_data=proj_data, ingest_api_agent=self.api) for proj_data in
                self.api.get_all(self.data['_links']['projects']['href'], 'projects')]

    def project(self):
        """ Assumes only one project """
        projects = self.projects()
        if len(projects) != 1:
            raise RuntimeError(f"Expect 1 project got {len(projects)}")
        return projects[0]

    @property
    def status(self):
        return self.data['submissionState']

    @property
    def uuid(self):
        return self.data['uuid']['uuid']

    def upload_credentials(self):
        """ Return upload area credentials or None if this envelope doesn't have an upload area yet """
        staging_details = self.data.get('stagingDetails', None)
        if staging_details and 'stagingAreaLocation' in staging_details:
            return staging_details.get('stagingAreaLocation', {}).get('value', None)
        return None

    def bundles(self):
        url = self.data['_links']['bundleManifests']['href']
        manifests = self.api.get_all(url, 'bundleManifests')
        return [manifest['bundleUuid'] for manifest in manifests]


class File:
    """
    Model an Ingest File entity
    """

    def __init__(self, file_data):
        self._data = file_data

    def __str__(self, prefix="", verbose=False):
        return (f"{prefix}    fileName {self.name}\n" +
                f"{prefix}    cloudUrl {self.cloud_url}\n")

    def __repr__(self):
        return json.dumps(self._data, indent=2)

    @property
    def name(self):
        return self._data['fileName']

    @property
    def cloud_url(self):
        return self._data['cloudUrl']



from urllib.parse import urlencode

import requests


class HateoasAgent:

    def __init__(self, api_url_base, auth_headers={}):
        self.api_url_base = api_url_base
        self.auth_headers = auth_headers

    """
    Get a collection resource.
    Iterates through all pages gathering results and returns a list.
    """
    def get_all(self, path_or_url, result_element_we_are_interested_in):
        results = []
        for page in self.iter_pages(path_or_url):
            results += page[result_element_we_are_interested_in]
        return results

    """
    Iterate through a collection using HATEOAS pagination, yielding pages.
    """
    def iter_pages(self, path_or_url, page_size=100, sort=None):
        url_params = {'size': page_size}
        if sort:
            url_params['sort'] = sort
        path_or_url += '?' + urlencode(url_params)

        while True:
            data = self.get(path_or_url)
            if '_embedded' not in data:
                break

            yield data['_embedded']

            if 'next' in data['_links']:
                path_or_url = data['_links']['next']['href']
            else:
                break

    """
    Get a singleton resource.
    """
    def get(self, path_or_url):
        if path_or_url.startswith('http'):
            url = path_or_url
        else:
            url = f"{self.api_url_base}{path_or_url}"

        response = requests.get(url, headers=self.auth_headers)

        if response.ok:
            return response.json()
        else:
            raise RuntimeError(f"GET {url} got {response}")



from .ingest_auth_agent import IngestAuthAgent
from .hateoas_agent import HateoasAgent


class IngestApiAgent(HateoasAgent):

    def __init__(self, deployment):
        self.deployment = deployment
        super().__init__(api_url_base=self._ingest_api_url(),
                         auth_headers=IngestAuthAgent().make_auth_header())

    def _ingest_api_url(self):
        if self.deployment == 'prod':
            return "http://api.ingest.data.humancellatlas.org"
        else:
            return f"http://api.ingest.{self.deployment}.data.humancellatlas.org"



from datetime import datetime

from sqlalchemy import create_engine, Column, Integer, String, DateTime, ForeignKey, JSON
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from termcolor import colored


Base = declarative_base()


class DbUploadArea(Base):
    __tablename__ = 'upload_area'
    id = Column(String(), primary_key=True)
    bucket_name = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    def __str__(self, prefix="", verbose=False):
        return colored(f"{prefix}Upload Area {self.id}\n", 'green') + \
            f"{prefix}    bucket_name: {self.bucket_name}\n" \
            f"{prefix}         Status: {self.status}\n" \
            f"{prefix}     created_at: {self.created_at}\n" \
            f"{prefix}     updated_at: {self.updated_at}\n"

    def show_associated(self, entities_to_show, prefix="", verbose=False):
        if 'files' in entities_to_show:
            for file in self.files:
                print(file.__str__(prefix=prefix, verbose=verbose))
                file.show_associated(entities_to_show, prefix="\t\t", verbose=verbose)


class DbFile(Base):
    __tablename__ = 'file'
    id = Column(String(), primary_key=True)
    upload_area_id = Column(String(), ForeignKey('upload_area.id'), nullable=False)
    name = Column(String(), nullable=False)
    size = Column(Integer(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    upload_area = relationship("DbUploadArea", back_populates="files")

    def __str__(self, prefix="", verbose=False):
        return colored(f"{prefix}File {self.id}\n", 'green') + \
            f"{prefix}    Upload Area ID: {self.upload_area_id}\n" \
            f"{prefix}              Name: {self.name}\n" \
            f"{prefix}              Size: {self.size}\n" \
            f"{prefix}        created_at: {self.created_at}\n" \
            f"{prefix}        updated_at: {self.updated_at}\n"

    def show_associated(self, entities_to_show, prefix="", verbose=False):
        if 'checksums' in entities_to_show:
            for checksum in self.checksums:
                print(checksum.__str__(prefix=prefix, verbose=verbose))
        if 'validations' in entities_to_show:
            for val in self.validations:
                print(val.__str__(prefix=prefix, verbose=verbose))
        if 'notifications' in entities_to_show:
            for notif in self.notifications:
                print(notif.__str__(prefix=prefix, verbose=verbose))


class DbChecksum(Base):
    __tablename__ = 'checksum'
    id = Column(String(), primary_key=True)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    job_id = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    checksums = Column(String(), nullable=False)
    checksum_started_at = Column(DateTime(), nullable=False)
    checksum_ended_at = Column(DateTime(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='checksums')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Checksum {self.id}\n", 'blue') + \
                 f"{prefix}                File ID: {self.file_id}\n" \
                 f"{prefix}                 Job ID: {self.job_id}\n" \
                 f"{prefix}                 Status: {self.status}\n" \
                 f"{prefix}    checksum_started_at: {self.checksum_started_at}\n" \
                 f"{prefix}      checksum_ended_at: {self.checksum_ended_at}\n" \
                 f"{prefix}             created_at: {self.created_at}\n" \
                 f"{prefix}             updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}\t          checksums: {self.checksums}\n"
        return output


class DbNotification(Base):
    __tablename__ = 'notification'
    id = Column(String(), primary_key=True)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    status = Column(String(), nullable=False)
    payload = Column(JSON(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='notifications')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Notification {self.id}\n", 'blue') + \
                 f"{prefix}       File ID: {self.file_id}\n" \
                 f"{prefix}        Status: {self.status}\n" \
                 f"{prefix}    created_at: {self.created_at}\n" \
                 f"{prefix}    updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}\t   payload: {self.payload}\n"
        return output


class DbValidation(Base):
    __tablename__ = 'validation'
    id = Column(String(), primary_key=True)
    file_id = Column(String(), ForeignKey('file.id'), nullable=False)
    job_id = Column(String(), nullable=False)
    status = Column(String(), nullable=False)
    results = Column(String(), nullable=False)
    validation_started_at = Column(DateTime(), nullable=False)
    validation_ended_at = Column(DateTime(), nullable=False)
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False, onupdate=datetime.utcnow)

    file = relationship("DbFile", back_populates='validations')

    def __str__(self, prefix="", verbose=False):
        output = colored(f"{prefix}Validation {self.id}\n", 'blue') + \
                 f"{prefix}                  File ID: {self.file_id}\n" \
                 f"{prefix}                   Job ID: {self.job_id}\n" \
                 f"{prefix}                   Status: {self.status}\n" \
                 f"{prefix}    validation_started_at: {self.validation_started_at}\n" \
                 f"{prefix}      validation_ended_at: {self.validation_ended_at}\n" \
                 f"{prefix}               created_at: {self.created_at}\n" \
                 f"{prefix}               updated_at: {self.updated_at}\n"
        if verbose:
            output = output + f"{prefix}\t              results: {self.results}\n"
        return output


DbUploadArea.files = relationship('DbFile', order_by=DbFile.id, back_populates='upload_area')
DbFile.checksums = relationship('DbChecksum', order_by=DbChecksum.created_at, back_populates='file')
DbFile.validations = relationship('DbValidation', order_by=DbValidation.created_at, back_populates='file')
DbFile.notifications = relationship('DbNotification', order_by=DbNotification.created_at, back_populates='file')


def init_db(database_uri):
    engine = create_engine(database_uri)
    Base.metadata.bind = engine
    db_session_maker = sessionmaker()
    db_session_maker.bind = engine
    return db_session_maker



import json
import os
from urllib.parse import urlencode

import requests


class DataStoreAgent:

    def __init__(self, deployment):
        self.deployment = deployment
        if self.deployment == 'prod':
            self.dss_url = "https://dss.data.humancellatlas.org/v1"
        else:
            self.dss_url = "https://dss.{deployment}.data.humancellatlas.org/v1".format(deployment=deployment)

    def search(self, query, replica='aws'):
        results = []
        url = f"{self.dss_url}/search"
        query_params = { 'replica': replica}
        query_json = {'es_query': query}
        for response in self.iter_pages(url, query_params=query_params, json_body=query_json):
            if 'results' not in response:
                raise RuntimeError(f"No results in response: {response}")
            results.extend(response['results'])
        return results

    def iter_pages(self, url, query_params={}, json_body={}, page_size=500):
        params = query_params.copy()
        params['per_page'] = page_size
        full_url = url + '?' + urlencode(params)

        while True:
            response = requests.post(full_url, json=json_body)
            yield response.json()

            link_header = response.headers.get('link', None)
            if link_header:
                next_link = link_header.split(';')[0]
                full_url = next_link.strip('<').strip('>')
            else:
                break

    def download_bundle(self, bundle_uuid, target_folder):
        print(f"Downloading bundle {bundle_uuid}:")
        manifest = self.bundle_manifest(bundle_uuid)
        bundle_folder = os.path.join(target_folder, bundle_uuid)
        try:
            os.makedirs(bundle_folder)
        except FileExistsError:
            pass

        for f in manifest['bundle']['files']:
            self.download_file(f['uuid'], save_as=os.path.join(bundle_folder, f['name']))
        return bundle_folder

    def bundle_manifest(self, bundle_uuid, replica='aws'):
        url = f"{self.dss_url}/bundles/{bundle_uuid}?replica={replica}"
        response = requests.get(url)
        assert response.ok
        assert response.headers['Content-type'] == 'application/json'
        return json.loads(response.content)

    def download_file(self, file_uuid, save_as, replica='aws'):
        url = f"{self.dss_url}/files/{file_uuid}?replica={replica}"
        print(f"Downloading file {file_uuid} to {save_as}")
        response = requests.get(url, stream=True)
        assert response.ok
        with open(save_as, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)



import glob
import os
from setuptools import setup, find_packages

install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

setup(name='dcp-diag',
      version='0.9.0',
      description='Data Coordination Platform diagnostic library and tools.',
      url='http://github.com/HumanCellAtlas/dcp-diag',
      author='Sam Pierson',
      author_email='spierson@chanzuckerberg.com',
      license='MIT',
      packages=find_packages(exclude=['tests']),
      scripts=glob.glob('scripts/*'),
      zip_safe=False,
      install_requires=install_requires,
      platforms=['MacOS X', 'Posix'],
      test_suite='test',
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Programming Language :: Python :: 3.6'
      ]
      )



from sqlalchemy.orm.exc import NoResultFound

from dcplib.config import Config

from .. import DcpDiagException
from .finder import Finder
from ..component_agents.upload_entities import DbUploadArea, DbFile, init_db


class UploadDbConfig(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(component_name='upload', secret_name='database', **kwargs)


class UploadFinder:

    """
    dcpdig @upload area=<uuid>
    dcpdig @upload file_id=<upload-area>/<filename>
    """

    name = 'upload'

    def __init__(self, deployment):
        self.deployment = deployment

    def find(self, expression):
        field_name, field_value = expression.split('=')
        db_session_maker = init_db(UploadDbConfig(deployment=self.deployment).database_uri)
        db = db_session_maker()

        if field_name == 'file_id':
            try:
                file = db.query(DbFile).filter(DbFile.id == field_value).one()
                return file
            except NoResultFound:
                raise DcpDiagException(f"No record of File \"{field_value}\""
                                       f" found in Upload's {self.deployment} database.")
        elif field_name == 'area':
            try:
                area = db.query(DbUploadArea).filter(DbUploadArea.id == field_value).one()
                return area
            except NoResultFound:
                raise DcpDiagException(f"No record of Upload Area \"{field_value}\""
                                       f" found in Upload's {self.deployment} database.")


Finder.register(UploadFinder)



import requests


class IngestAuthAgent:
    def __init__(self,
                 url="https://danielvaughan.eu.auth0.com/oauth/token",
                 client_id="Zdsog4nDAnhQ99yiKwMQWAPc2qUDlR99",
                 client_secret="t-OAE-GQk_nZZtWn-QQezJxDsLXmU7VSzlAh9cKW5vb87i90qlXGTvVNAjfT9weF",
                 audience="http://localhost:8080",
                 grant_type="client_credentials"):
        """This class controls the authentication actions with Ingest Service, including retrieving the token,
            store the token and make authenticated headers. Note: The parameters and credentials here are
            meant to be hard coded, the authentication is purely for identifying a user it doesn't give any permissions.

        :param str url: The url to the Auth0 domain oauth endpoint.
        :param str client_id: The value of the Client ID field of the Non Interactive Client of Auth0.
        :param str client_secret: The value of the Client Secret field of the Non Interactive Client of Auth0.
        :param str audience: The value of the Identifier field of the Auth0 Management API.
        :param str grant_type: Denotes which OAuth 2.0 flow you want to run. e.g. client_credentials
        """
        self.url = url
        self.client_id = client_id
        self.client_secret = client_secret
        self.audience = audience
        self.grant_type = grant_type
        self.auth_token = self._get_auth_token()

    def _get_auth_token(self):
        """Request and get the access token for a trusted client from Auth0.

        :return dict auth_token: JSON response of the signed JWT (JSON Web Token), with when it expires (24h by
            default), the scopes granted, and the token type.
        """
        url = self.url
        headers = {
            "content-type": "application/json"
        }
        payload = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "audience": self.audience,
            "grant_type": self.grant_type
        }
        response = requests.post(url=url, headers=headers, json=payload)
        response.raise_for_status()
        auth_token = response.json()
        return auth_token

    def make_auth_header(self):
        """Make the authorization headers to communicate with endpoints which implement Auth0 authentication API.

        :return dict headers: A header with necessary token information to talk to Auth0 authentication required
            endpoints.
        """
        token_type = self.auth_token['token_type']
        access_token = self.auth_token['access_token']

        headers = {
            "Authorization": f"{token_type} {access_token}"
        }
        return headers
